@@ -1,12 +1,24 @@
 use super::{data::DataSerializer, sigma_byte_writer::SigmaByteWrite};
 use crate::mir::constant::Constant;
+use crate::serialization::types::TypeCode;
 use crate::serialization::{
     sigma_byte_reader::SigmaByteRead, SerializationError, SigmaSerializable,
 };
 use crate::types::stype::SType;
-
 use std::io;
 
+impl Constant {
+    /// Parse constant from byte stream. This function should be used instead of
+    /// `sigma_parse` when type code is already read for look-ahead
+    pub fn parse_with_type_code<R: SigmaByteRead>(
+        r: &mut R,
+        t_code: TypeCode,
+    ) -> Result<Self, SerializationError> {
+        let tpe = SType::parse_with_type_code(r, t_code)?;
+        let v = DataSerializer::sigma_parse(&tpe, r)?;
+        Ok(Constant { tpe, v })
+    }
+}
 impl SigmaSerializable for Constant {
     fn sigma_serialize<W: SigmaByteWrite>(&self, w: &mut W) -> Result<(), io::Error> {
         self.tpe.sigma_serialize(w)?;
@@ -15,9 +27,8 @@ impl SigmaSerializable for Constant {
 
     fn sigma_parse<R: SigmaByteRead>(r: &mut R) -> Result<Self, SerializationError> {
         // for reference see http://github.com/ScorexFoundation/sigmastate-interpreter/blob/25251c1313b0131835f92099f02cef8a5d932b5e/sigmastate/src/main/scala/sigmastate/serialization/DataSerializer.scala#L84-L84
-        let tpe = SType::sigma_parse(r)?;
-        let v = DataSerializer::sigma_parse(&tpe, r)?;
-        Ok(Constant { tpe, v })
+        let t_code = TypeCode::sigma_parse(r)?;
+        Self::parse_with_type_code(r, t_code)
     }
 }
 
