@@ -61,8 +61,116 @@ use crate::serialization::{
     sigma_byte_reader::SigmaByteRead, SerializationError, SigmaSerializable,
 };
 
+use crate::serialization::types::TypeCode;
 use std::io;
 
+impl Expr {
+    /// Parse expression from byte stream. This function should be used instead of
+    /// `sigma_parse` when tag byte is already read for look-ahead
+    pub fn parse_with_tag<R: SigmaByteRead>(
+        r: &mut R,
+        tag: u8,
+    ) -> Result<Self, SerializationError> {
+        let res = if tag <= OpCode::LAST_CONSTANT_CODE.value() {
+            let t_code = TypeCode::parse(tag)?;
+            let constant = Constant::parse_with_type_code(r, t_code)?;
+            Ok(Expr::Const(constant))
+        } else {
+            let op_code = OpCode::parse(tag);
+            match op_code {
+                OpCode::FOLD => Ok(Fold::sigma_parse(r)?.into()),
+                ConstantPlaceholder::OP_CODE => {
+                    let cp = ConstantPlaceholder::sigma_parse(r)?;
+                    if r.substitute_placeholders() {
+                        // ConstantPlaceholder itself can be created only if a corresponding
+                        // constant is in the constant_store, thus unwrap() is safe here
+                        let c = r.constant_store().get(cp.id).unwrap();
+                        Ok(Expr::Const(c.clone()))
+                    } else {
+                        Ok(Expr::ConstPlaceholder(cp))
+                    }
+                }
+                OpCode::HEIGHT => Ok(Expr::GlobalVars(GlobalVars::Height)),
+                OpCode::SELF_BOX => Ok(Expr::GlobalVars(GlobalVars::SelfBox)),
+                OpCode::INPUTS => Ok(Expr::GlobalVars(GlobalVars::Inputs)),
+                OpCode::OUTPUTS => Ok(Expr::GlobalVars(GlobalVars::Outputs)),
+                OpCode::MINER_PUBKEY => Ok(Expr::GlobalVars(GlobalVars::MinerPubKey)),
+                OpCode::GLOBAL => Ok(Expr::Global),
+                OpCode::PROPERTY_CALL => Ok(Expr::ProperyCall(PropertyCall::sigma_parse(r)?)),
+                OpCode::METHOD_CALL => Ok(Expr::MethodCall(MethodCall::sigma_parse(r)?)),
+                OpCode::CONTEXT => Ok(Expr::Context),
+                OptionGet::OP_CODE => Ok(OptionGet::sigma_parse(r)?.into()),
+                OptionIsDefined::OP_CODE => Ok(OptionIsDefined::sigma_parse(r)?.into()),
+                OptionGetOrElse::OP_CODE => Ok(OptionGetOrElse::sigma_parse(r)?.into()),
+                ExtractRegisterAs::OP_CODE => Ok(ExtractRegisterAs::sigma_parse(r)?.into()),
+                ExtractScriptBytes::OP_CODE => Ok(ExtractScriptBytes::sigma_parse(r)?.into()),
+                ExtractCreationInfo::OP_CODE => Ok(ExtractCreationInfo::sigma_parse(r)?.into()),
+                ExtractId::OP_CODE => Ok(ExtractId::sigma_parse(r)?.into()),
+                OpCode::EQ => Ok(bin_op_sigma_parse(RelationOp::Eq.into(), r)?),
+                OpCode::NEQ => Ok(bin_op_sigma_parse(RelationOp::NEq.into(), r)?),
+                Negation::OP_CODE => Ok(Negation::sigma_parse(r)?.into()),
+                OpCode::LOGICAL_NOT => Ok(LogicalNot::sigma_parse(r)?.into()),
+                OpCode::BIN_AND => Ok(bin_op_sigma_parse(RelationOp::And.into(), r)?),
+                OpCode::BIN_OR => Ok(bin_op_sigma_parse(RelationOp::Or.into(), r)?),
+                OpCode::GT => Ok(bin_op_sigma_parse(RelationOp::Gt.into(), r)?),
+                OpCode::LT => Ok(bin_op_sigma_parse(RelationOp::Lt.into(), r)?),
+                OpCode::GE => Ok(bin_op_sigma_parse(RelationOp::Ge.into(), r)?),
+                OpCode::LE => Ok(bin_op_sigma_parse(RelationOp::Le.into(), r)?),
+                OpCode::PLUS => Ok(bin_op_sigma_parse(ArithOp::Plus.into(), r)?),
+                OpCode::MINUS => Ok(bin_op_sigma_parse(ArithOp::Minus.into(), r)?),
+                OpCode::MULTIPLY => Ok(bin_op_sigma_parse(ArithOp::Multiply.into(), r)?),
+                OpCode::DIVISION => Ok(bin_op_sigma_parse(ArithOp::Divide.into(), r)?),
+                OpCode::MAX => Ok(bin_op_sigma_parse(ArithOp::Max.into(), r)?),
+                OpCode::MIN => Ok(bin_op_sigma_parse(ArithOp::Min.into(), r)?),
+                OpCode::BIT_OR => Ok(bin_op_sigma_parse(ArithOp::BitOr.into(), r)?),
+                OpCode::BIT_AND => Ok(bin_op_sigma_parse(ArithOp::BitAnd.into(), r)?),
+                OpCode::BIT_XOR => Ok(bin_op_sigma_parse(ArithOp::BitXor.into(), r)?),
+                OpCode::BLOCK_VALUE => Ok(Expr::BlockValue(BlockValue::sigma_parse(r)?)),
+                OpCode::FUNC_VALUE => Ok(Expr::FuncValue(FuncValue::sigma_parse(r)?)),
+                OpCode::APPLY => Ok(Expr::Apply(Apply::sigma_parse(r)?)),
+                OpCode::VAL_DEF => Ok(Expr::ValDef(ValDef::sigma_parse(r)?)),
+                OpCode::VAL_USE => Ok(Expr::ValUse(ValUse::sigma_parse(r)?)),
+                ExtractAmount::OP_CODE => Ok(Expr::ExtractAmount(ExtractAmount::sigma_parse(r)?)),
+                OpCode::SELECT_FIELD => Ok(Expr::SelectField(SelectField::sigma_parse(r)?)),
+                OpCode::CALC_BLAKE2B256 => Ok(CalcBlake2b256::sigma_parse(r)?.into()),
+                And::OP_CODE => Ok(And::sigma_parse(r)?.into()),
+                Or::OP_CODE => Ok(Or::sigma_parse(r)?.into()),
+                Atleast::OP_CODE => Ok(Atleast::sigma_parse(r)?.into()),
+                OpCode::COLL => Ok(coll_sigma_parse(r)?.into()),
+                OpCode::COLL_OF_BOOL_CONST => Ok(bool_const_coll_sigma_parse(r)?.into()),
+                Map::OP_CODE => Ok(Map::sigma_parse(r)?.into()),
+                Filter::OP_CODE => Ok(Filter::sigma_parse(r)?.into()),
+                Exists::OP_CODE => Ok(Exists::sigma_parse(r)?.into()),
+                ForAll::OP_CODE => Ok(ForAll::sigma_parse(r)?.into()),
+                BoolToSigmaProp::OP_CODE => Ok(BoolToSigmaProp::sigma_parse(r)?.into()),
+                Upcast::OP_CODE => Ok(Upcast::sigma_parse(r)?.into()),
+                If::OP_CODE => Ok(If::sigma_parse(r)?.into()),
+                ByIndex::OP_CODE => Ok(ByIndex::sigma_parse(r)?.into()),
+                SizeOf::OP_CODE => Ok(SizeOf::sigma_parse(r)?.into()),
+                CreateProveDlog::OP_CODE => Ok(CreateProveDlog::sigma_parse(r)?.into()),
+                CreateProveDhTuple::OP_CODE => Ok(CreateProveDhTuple::sigma_parse(r)?.into()),
+                SigmaPropBytes::OP_CODE => Ok(SigmaPropBytes::sigma_parse(r)?.into()),
+                Tuple::OP_CODE => Ok(Tuple::sigma_parse(r)?.into()),
+                DecodePoint::OP_CODE => Ok(DecodePoint::sigma_parse(r)?.into()),
+                ByteArrayToLong::OP_CODE => Ok(ByteArrayToLong::sigma_parse(r)?.into()),
+                ByteArrayToBigInt::OP_CODE => Ok(ByteArrayToBigInt::sigma_parse(r)?.into()),
+                LongToByteArray::OP_CODE => Ok(LongToByteArray::sigma_parse(r)?.into()),
+                SigmaAnd::OP_CODE => Ok(SigmaAnd::sigma_parse(r)?.into()),
+                SigmaOr::OP_CODE => Ok(SigmaOr::sigma_parse(r)?.into()),
+                GetVar::OP_CODE => Ok(GetVar::sigma_parse(r)?.into()),
+                DeserializeRegister::OP_CODE => Ok(DeserializeRegister::sigma_parse(r)?.into()),
+                DeserializeContext::OP_CODE => Ok(DeserializeContext::sigma_parse(r)?.into()),
+                o => Err(SerializationError::NotImplementedOpCode(format!(
+                    "{0}(shift {1})",
+                    o.value(),
+                    o.shift()
+                ))),
+            }
+        };
+        res
+    }
+}
+
 impl SigmaSerializable for Expr {
     fn sigma_serialize<W: SigmaByteWrite>(&self, w: &mut W) -> Result<(), io::Error> {
         match self {
@@ -136,107 +244,8 @@ impl SigmaSerializable for Expr {
     }
 
     fn sigma_parse<R: SigmaByteRead>(r: &mut R) -> Result<Self, SerializationError> {
-        let first_byte = match r.peek_u8() {
-            Ok(b) => Ok(b),
-            Err(_) => {
-                let res = r.get_u8(); // get(consume) the error
-                assert!(res.is_err());
-                res
-            }
-        }?;
-        let res = if first_byte <= OpCode::LAST_CONSTANT_CODE.value() {
-            let constant = Constant::sigma_parse(r)?;
-            Ok(Expr::Const(constant))
-        } else {
-            let op_code = OpCode::sigma_parse(r)?;
-            match op_code {
-                OpCode::FOLD => Ok(Fold::sigma_parse(r)?.into()),
-                ConstantPlaceholder::OP_CODE => {
-                    let cp = ConstantPlaceholder::sigma_parse(r)?;
-                    if r.substitute_placeholders() {
-                        // ConstantPlaceholder itself can be created only if a corresponding
-                        // constant is in the constant_store, thus unwrap() is safe here
-                        let c = r.constant_store().get(cp.id).unwrap();
-                        Ok(Expr::Const(c.clone()))
-                    } else {
-                        Ok(Expr::ConstPlaceholder(cp))
-                    }
-                }
-                OpCode::HEIGHT => Ok(Expr::GlobalVars(GlobalVars::Height)),
-                OpCode::SELF_BOX => Ok(Expr::GlobalVars(GlobalVars::SelfBox)),
-                OpCode::INPUTS => Ok(Expr::GlobalVars(GlobalVars::Inputs)),
-                OpCode::OUTPUTS => Ok(Expr::GlobalVars(GlobalVars::Outputs)),
-                OpCode::MINER_PUBKEY => Ok(Expr::GlobalVars(GlobalVars::MinerPubKey)),
-                OpCode::GLOBAL => Ok(Expr::Global),
-                OpCode::PROPERTY_CALL => Ok(Expr::ProperyCall(PropertyCall::sigma_parse(r)?)),
-                OpCode::METHOD_CALL => Ok(Expr::MethodCall(MethodCall::sigma_parse(r)?)),
-                OpCode::CONTEXT => Ok(Expr::Context),
-                OptionGet::OP_CODE => Ok(OptionGet::sigma_parse(r)?.into()),
-                OptionIsDefined::OP_CODE => Ok(OptionIsDefined::sigma_parse(r)?.into()),
-                OptionGetOrElse::OP_CODE => Ok(OptionGetOrElse::sigma_parse(r)?.into()),
-                ExtractRegisterAs::OP_CODE => Ok(ExtractRegisterAs::sigma_parse(r)?.into()),
-                ExtractScriptBytes::OP_CODE => Ok(ExtractScriptBytes::sigma_parse(r)?.into()),
-                ExtractCreationInfo::OP_CODE => Ok(ExtractCreationInfo::sigma_parse(r)?.into()),
-                ExtractId::OP_CODE => Ok(ExtractId::sigma_parse(r)?.into()),
-                OpCode::EQ => Ok(bin_op_sigma_parse(RelationOp::Eq.into(), r)?),
-                OpCode::NEQ => Ok(bin_op_sigma_parse(RelationOp::NEq.into(), r)?),
-                Negation::OP_CODE => Ok(Negation::sigma_parse(r)?.into()),
-                OpCode::LOGICAL_NOT => Ok(LogicalNot::sigma_parse(r)?.into()),
-                OpCode::BIN_AND => Ok(bin_op_sigma_parse(RelationOp::And.into(), r)?),
-                OpCode::BIN_OR => Ok(bin_op_sigma_parse(RelationOp::Or.into(), r)?),
-                OpCode::GT => Ok(bin_op_sigma_parse(RelationOp::Gt.into(), r)?),
-                OpCode::LT => Ok(bin_op_sigma_parse(RelationOp::Lt.into(), r)?),
-                OpCode::GE => Ok(bin_op_sigma_parse(RelationOp::Ge.into(), r)?),
-                OpCode::LE => Ok(bin_op_sigma_parse(RelationOp::Le.into(), r)?),
-                OpCode::PLUS => Ok(bin_op_sigma_parse(ArithOp::Plus.into(), r)?),
-                OpCode::MINUS => Ok(bin_op_sigma_parse(ArithOp::Minus.into(), r)?),
-                OpCode::MULTIPLY => Ok(bin_op_sigma_parse(ArithOp::Multiply.into(), r)?),
-                OpCode::DIVISION => Ok(bin_op_sigma_parse(ArithOp::Divide.into(), r)?),
-                OpCode::MAX => Ok(bin_op_sigma_parse(ArithOp::Max.into(), r)?),
-                OpCode::MIN => Ok(bin_op_sigma_parse(ArithOp::Min.into(), r)?),
-                OpCode::BLOCK_VALUE => Ok(Expr::BlockValue(BlockValue::sigma_parse(r)?)),
-                OpCode::FUNC_VALUE => Ok(Expr::FuncValue(FuncValue::sigma_parse(r)?)),
-                OpCode::APPLY => Ok(Expr::Apply(Apply::sigma_parse(r)?)),
-                OpCode::VAL_DEF => Ok(Expr::ValDef(ValDef::sigma_parse(r)?)),
-                OpCode::VAL_USE => Ok(Expr::ValUse(ValUse::sigma_parse(r)?)),
-                ExtractAmount::OP_CODE => Ok(Expr::ExtractAmount(ExtractAmount::sigma_parse(r)?)),
-                OpCode::SELECT_FIELD => Ok(Expr::SelectField(SelectField::sigma_parse(r)?)),
-                OpCode::CALC_BLAKE2B256 => Ok(CalcBlake2b256::sigma_parse(r)?.into()),
-                And::OP_CODE => Ok(And::sigma_parse(r)?.into()),
-                Or::OP_CODE => Ok(Or::sigma_parse(r)?.into()),
-                Atleast::OP_CODE => Ok(Atleast::sigma_parse(r)?.into()),
-                OpCode::COLL => Ok(coll_sigma_parse(r)?.into()),
-                OpCode::COLL_OF_BOOL_CONST => Ok(bool_const_coll_sigma_parse(r)?.into()),
-                Map::OP_CODE => Ok(Map::sigma_parse(r)?.into()),
-                Filter::OP_CODE => Ok(Filter::sigma_parse(r)?.into()),
-                Exists::OP_CODE => Ok(Exists::sigma_parse(r)?.into()),
-                ForAll::OP_CODE => Ok(ForAll::sigma_parse(r)?.into()),
-                BoolToSigmaProp::OP_CODE => Ok(BoolToSigmaProp::sigma_parse(r)?.into()),
-                Upcast::OP_CODE => Ok(Upcast::sigma_parse(r)?.into()),
-                If::OP_CODE => Ok(If::sigma_parse(r)?.into()),
-                ByIndex::OP_CODE => Ok(ByIndex::sigma_parse(r)?.into()),
-                SizeOf::OP_CODE => Ok(SizeOf::sigma_parse(r)?.into()),
-                CreateProveDlog::OP_CODE => Ok(CreateProveDlog::sigma_parse(r)?.into()),
-                CreateProveDhTuple::OP_CODE => Ok(CreateProveDhTuple::sigma_parse(r)?.into()),
-                SigmaPropBytes::OP_CODE => Ok(SigmaPropBytes::sigma_parse(r)?.into()),
-                Tuple::OP_CODE => Ok(Tuple::sigma_parse(r)?.into()),
-                DecodePoint::OP_CODE => Ok(DecodePoint::sigma_parse(r)?.into()),
-                ByteArrayToLong::OP_CODE => Ok(ByteArrayToLong::sigma_parse(r)?.into()),
-                ByteArrayToBigInt::OP_CODE => Ok(ByteArrayToBigInt::sigma_parse(r)?.into()),
-                LongToByteArray::OP_CODE => Ok(LongToByteArray::sigma_parse(r)?.into()),
-                SigmaAnd::OP_CODE => Ok(SigmaAnd::sigma_parse(r)?.into()),
-                SigmaOr::OP_CODE => Ok(SigmaOr::sigma_parse(r)?.into()),
-                GetVar::OP_CODE => Ok(GetVar::sigma_parse(r)?.into()),
-                DeserializeRegister::OP_CODE => Ok(DeserializeRegister::sigma_parse(r)?.into()),
-                DeserializeContext::OP_CODE => Ok(DeserializeContext::sigma_parse(r)?.into()),
-                o => Err(SerializationError::NotImplementedOpCode(format!(
-                    "{0}(shift {1})",
-                    o.value(),
-                    o.shift()
-                ))),
-            }
-        };
-        res
+        let tag = r.get_u8()?;
+        Self::parse_with_tag(r, tag)
     }
 }
 
