@@ -10,9 +10,11 @@ use super::{
 };
 use crate::serialization::types::TypeCode;
 use crate::types::smethod::MethodId;
+use bounded_vec::BoundedVec;
 use bounded_vec::BoundedVecOutOfBounds;
 use io::Cursor;
-use sigma_ser::{peekable_reader::PeekableReader, vlq_encode};
+use sigma_ser::vlq_encode;
+use std::convert::TryInto;
 use std::io;
 use thiserror::Error;
 
@@ -30,9 +32,9 @@ pub enum SerializationError {
     InvalidTypePrefix,
     /// Failed to decode VLQ
     #[error("vlq encode error: {0}")]
-    VlqEncode(vlq_encode::VlqEncodingError),
+    VlqEncode(#[from] vlq_encode::VlqEncodingError),
     /// IO fail (EOF, etc.)
-    #[error("io error")]
+    #[error("IO error: {0}")]
     Io(String),
     /// Misc fail
     #[error("misc error")]
@@ -60,12 +62,6 @@ pub enum SerializationError {
     UnknownMethodId(MethodId, TypeCode),
 }
 
-impl From<vlq_encode::VlqEncodingError> for SerializationError {
-    fn from(error: vlq_encode::VlqEncodingError) -> Self {
-        SerializationError::VlqEncode(error)
-    }
-}
-
 impl From<io::Error> for SerializationError {
     fn from(error: io::Error) -> Self {
         SerializationError::Io(error.to_string())
@@ -126,8 +122,7 @@ pub trait SigmaSerializable: Sized {
     /// Parse `self` from the bytes
     fn sigma_parse_bytes(bytes: &[u8]) -> Result<Self, SerializationError> {
         let cursor = Cursor::new(bytes);
-        let pr = PeekableReader::new(cursor);
-        let mut sr = SigmaByteReader::new(pr, ConstantStore::empty());
+        let mut sr = SigmaByteReader::new(cursor, ConstantStore::empty());
         Self::sigma_parse(&mut sr)
     }
 }
@@ -148,6 +143,18 @@ impl<T: SigmaSerializable> SigmaSerializable for Vec<T> {
     }
 }
 
+impl<T: SigmaSerializable, const L: usize, const U: usize> SigmaSerializable
+    for BoundedVec<T, L, U>
+{
+    fn sigma_serialize<W: SigmaByteWrite>(&self, w: &mut W) -> Result<(), io::Error> {
+        self.as_vec().sigma_serialize(w)
+    }
+
+    fn sigma_parse<R: SigmaByteRead>(r: &mut R) -> Result<Self, SerializationError> {
+        Ok(Vec::<T>::sigma_parse(r)?.try_into()?)
+    }
+}
+
 impl<T: SigmaSerializable> SigmaSerializable for Option<Box<T>> {
     fn sigma_serialize<W: SigmaByteWrite>(&self, w: &mut W) -> Result<(), io::Error> {
         match self {
@@ -176,7 +183,6 @@ pub fn sigma_serialize_roundtrip<T: SigmaSerializable>(v: &T) -> T {
     let mut w = SigmaByteWriter::new(&mut data, None);
     v.sigma_serialize(&mut w).expect("serialization failed");
     let cursor = Cursor::new(&mut data[..]);
-    let pr = PeekableReader::new(cursor);
-    let mut sr = SigmaByteReader::new(pr, ConstantStore::empty());
+    let mut sr = SigmaByteReader::new(cursor, ConstantStore::empty());
     T::sigma_parse(&mut sr).expect("parse failed")
 }
